"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const rx = require("rxjs");
class Upset {
    static create(etcd) {
        return new Upset(etcd);
    }
    constructor(etcd) {
        this.etcd = etcd;
    }
    runApply(inp, apply, obs, uuid) {
        const sub = new rx.Subject();
        sub.subscribe((toSet) => {
            this.etcd.cfg.log.debug('runApply:next', uuid, toSet);
            obs.next(toSet);
        }, err => {
            this.etcd.cfg.log.debug('runApply:error', uuid);
            obs.error(err);
        }, () => {
            this.etcd.cfg.log.debug('runApply:complete', uuid);
            obs.complete();
        });
        apply(inp, sub);
    }
    updateIt(prevValue, key, toSet, apply, obs, uuid) {
        const options = {};
        if (!prevValue) {
            options.prevExist = false;
        }
        else {
            options.prevValue = JSON.stringify(prevValue);
        }
        this.etcd.setJson(key, toSet, options).subscribe((er) => {
            if (er.isErr()) {
                if (er.err.etcErr.errorCode == 105) {
                    this.etcd.cfg.log.debug('updateIt:retry', uuid);
                    this.doIt(key, apply, obs, uuid);
                }
                else {
                    this.etcd.cfg.log.debug('updateIt:next:error', uuid, er);
                    obs.error(er);
                }
                return;
            }
            this.etcd.cfg.log.debug('updateIt:next:', uuid, er.node.modifiedIndex);
            obs.next(toSet);
            obs.complete();
        }, err => {
            this.etcd.cfg.log.debug('updateIt:error:', uuid, err);
            obs.error(err);
        }, () => {
            /* this.etcd.cfg.log.info('updateIt:complete:', uuid); */
        });
    }
    doIt(key, apply, obs, uuid) {
        // read key
        const applyStep = new rx.Subject();
        let applyNextReceived = false; // very unhappy!
        let prevValue = null;
        applyStep.subscribe((toSet) => {
            applyNextReceived = true;
            this.etcd.cfg.log.debug('applyStep:updateIt:', uuid, toSet);
            this.updateIt(prevValue, key, toSet, apply, obs, uuid);
        }, err => {
            this.etcd.cfg.log.debug('applyStep:error:', uuid, err);
            obs.error(err);
        }, () => {
            this.etcd.cfg.log.debug('applyStep:complete:', uuid, applyNextReceived);
            if (!applyNextReceived) {
                obs.complete();
            }
        });
        this.etcd.getJson(key).subscribe((_obs) => {
            if (_obs.isErr()) {
                if (_obs.err.err.etcErr.errorCode == 100) {
                    this.etcd.cfg.log.debug('getJson:next:notfound:', uuid, key);
                    prevValue = null;
                    this.runApply(null, apply, applyStep, uuid);
                }
                else {
                    this.etcd.cfg.log.error('getJson:next:error:', uuid, key, _obs.err.err.etcErr.errorCode);
                    applyStep.error(_obs);
                }
            }
            else {
                this.etcd.cfg.log.debug('getJson:next:found:', uuid, key);
                prevValue = _obs.value;
                this.runApply(_obs.value, apply, applyStep, uuid);
            }
        }, (err) => {
            applyStep.error(err);
        }, () => {
            /* */
        });
    }
    upSet(key, apply) {
        let once = true;
        return rx.Observable.create((obs) => {
            if (once) {
                once = false;
                this.doIt(key, apply, obs, Math.random());
            }
            else {
                this.etcd.cfg.log.error('upSet:double:once');
            }
        });
    }
}
exports.Upset = Upset;
exports.default = Upset;
//# sourceMappingURL=upset.js.map